#include <stdio.h>
#include <time.h>
#include <string.h>
#include <syslog.h>
#include <stdarg.h>
#define NARGS 20

void myx_syslog(int pri,const char *format, ...);
void myx_syslogdf(int pri,char *fname);
static int lparse(char *line,char *argv[],int maxargs,char delc);
static char *gethsmpin();

static char tfile[512];

int main(int argc,char *argv[])
{
  FILE *fp,*fp2;
  char buf[512],lbuf[512],now[50],*args[NARGS];
  time_t curtime;
  struct tm *tnow,_tnow;

  openlog(argv[0],LOG_PID,LOG_AUTHPRIV);

  sprintf(tfile,"%s.tmp",argv[0]);

  time(&curtime);
  tnow = &_tnow;
  memcpy(tnow,gmtime(&curtime),sizeof(struct tm));
  sprintf(now,"%04u%02u%02u%02u%02u%02u",tnow->tm_year+1900,tnow->tm_mon + 1,tnow->tm_mday,tnow->tm_hour,tnow->tm_min,tnow->tm_sec);

  myx_syslog(LOG_NOTICE,"All Key backup started\n");
  if((fp=fopen("keyindex","r")) == NULL) {
    myx_syslog(LOG_ERR,"Could not open keyindex file\n");
    return -1;
  }
  while(fgets(buf,sizeof(buf),fp)) {
    if(buf[0] == '#') continue;
    if(lparse(buf,args,NARGS,' ') < 8) continue;
    if(strcmp(args[7],now) < 0) continue;
    sprintf(lbuf,"keydir/K%s.+%s+%s.private",args[0],args[2],args[3]);
    if((fp2=fopen(lbuf,"r"))) {
      if(fgets(lbuf,sizeof(lbuf),fp2)) {
	lbuf[strlen(lbuf) - 1] = '\0';
	if(strcmp(lbuf,"Private-key-format: xxx") == 0) {
	  sprintf(lbuf,"./pkcs11-backup %s -pK%s > K%s.hsm",gethsmpin(),args[3],args[3]);
	  if(system(lbuf)) {
	    myx_syslog(LOG_ERR,"Could not execute %s\n",lbuf);
	    return -1;
	  }
	}
      }
      fclose(fp2);
    }
    sprintf(lbuf,"tar vrf keyall.tar keydir/K%s.+%s+%s.* 2>&1 > %s",args[0],args[2],args[3],tfile);
    if(system(lbuf)) {
      myx_syslog(LOG_ERR,"Could not execute %s\n",lbuf);
    }
    myx_syslogdf(LOG_INFO,tfile);
  }
  fclose(fp);
  sprintf(lbuf,"tar vrf keyall.tar keyindex *.hsm HSM/*.db fob 2>&1 > %s; rm -f *.hsm",tfile);
  if(system(lbuf)) {
    myx_syslog(LOG_ERR,"Could not execute %s\n",lbuf);
  }
  myx_syslogdf(LOG_INFO,tfile);

  myx_syslog(LOG_NOTICE,"Encrypting keyall.tar with HSM...\n");

  sprintf(lbuf,"./pkcs11-encrypt %s -S 0 -e -i keyall.tar -o keyall.tar.hsm 2>&1 > %s",gethsmpin(),tfile);
  if(system(lbuf)) {
    myx_syslog(LOG_ERR,"Could not execute %s\n",lbuf);
  }
  myx_syslogdf(LOG_INFO,tfile);

  unlink("keyall.tar");

  myx_syslog(LOG_NOTICE,"Appended and encrypted above files to keyall.tar.hsm\n");
  unlink(tfile);

  return 0;
}
static char *gethsmpin()
{
  static char *hsmpin;
  char lbuf[80];

  if(hsmpin) return hsmpin;
  myx_syslog(LOG_INFO,"Please enter PIN: ");
  sprintf(lbuf,"-P");
  if(fgets(&lbuf[2],sizeof(lbuf)-3,stdin)) {
    lbuf[sizeof(lbuf)-3] = '\0';
    lbuf[strlen(lbuf) - 1] = '\0';
    hsmpin = strdup(lbuf);
  } else {
    myx_syslog(LOG_ERR,"PIN input error\n");
    return "";
  }
  return hsmpin;
}
void myx_syslog(int pri,const char *format, ...)
{
  va_list args;

  va_start(args,format);
  vfprintf(stdout,format,args);
  vsyslog(pri,format,args);
  va_end(args);
}
void myx_syslogdf(int pri,char *fname)
{
  FILE *fp;
  char lbuf[256];

  if((fp=fopen(fname,"r")) == NULL) return;
  while(fgets(lbuf,sizeof(lbuf),fp)) myx_syslog(pri,"%s",lbuf);
  fclose(fp);
}
static int lparse(char *line,char *argv[],int maxargs,char delc)
{
  char *cp;
  int argc,qflag;

  if((cp = strchr(line,'\r')) != (char *)0) *cp = '\0';
  if((cp = strchr(line,'\n')) != (char *)0) *cp = '\0';

  for(argc=0;argc<maxargs;argc++) argv[argc] = (char *)0;

  for(argc=0;argc<maxargs;) {
    qflag = 0;
    while(*line == ' ' || *line == '\t') line++; /* whitespace */
    if(*line == '\0') break; /* done */
    if(*line == '"') { line++; qflag = 1; } /* quote */
    argv[argc++] = line;
    if(qflag) {                         /* quote */
      if((line = strchr(line,'"')) == (char *)0) return -1; /*error*/
      *line++ = '\0';
    } else {
      for(cp=line;*cp;cp++) {
        if(*cp == ' ' || *cp == '\t' || *cp == delc) break;
      }
      if(*cp) *cp++ = '\0'; /* non-zero */
      line = cp;
    }
  }
  return argc;
}
